#include "Matrix.hpp"

void CMatrix::identity() {
	m[0][0] = 1.0f; m[0][1] = 0.0f; m[0][2] = 0.0f; m[0][3] = 0.0f;
	m[1][0] = 0.0f; m[1][1] = 1.0f; m[1][2] = 0.0f; m[1][3] = 0.0f;
	m[2][0] = 0.0f; m[2][1] = 0.0f; m[2][2] = 1.0f; m[2][3] = 0.0f;
	m[3][0] = 0.0f; m[3][1] = 0.0f; m[3][2] = 0.0f; m[3][3] = 1.0f;
}

void CMatrix::multiply(const CMatrix& d) {
	float r[4][4];

	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++) {
			r[i][j] = 0.0f;
			for (int k = 0; k < 4; k++) {
				r[i][j] += m[i][k] * d.m[k][j];
			}
		}
	}
	memcpy(&m[0][0], &r[0][0], sizeof(m));
}

CVector CMatrix::multiply(float x, float y, float z) {
	CVector v;
	v.x = x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0];
	v.y = x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1];
	v.z = x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2];
	return v;
}

void CMatrix::translate(float x, float y, float z) {
	CMatrix d;
	d.m[3][0] = x; d.m[3][1] = y; d.m[3][2] = z;
	multiply(d);
}

void CMatrix::scale(float s) {
	CMatrix d;
	d.m[0][0] = s; d.m[1][1] = s; d.m[2][2] = s;
	multiply(d);
}

void CMatrix::rotateX(float angle) {
	float sin = sinf(angle);
	float cos = cosf(angle);
	CMatrix d;
	d.m[1][1] = cos; d.m[1][2] = sin;
	d.m[2][1] = -sin; d.m[2][2] = cos;
	multiply(d);
}

void CMatrix::rotateY(float angle) {
	float sin = sinf(angle);
	float cos = cosf(angle);
	CMatrix d;
	d.m[0][0] = cos; d.m[0][2] = -sin;
	d.m[2][0] = sin; d.m[2][2] = cos;
	multiply(d);
}

void CMatrix::rotateZ(float angle) {
	float sin = sinf(angle);
	float cos = cosf(angle);
	CMatrix d;
	d.m[0][0] = cos; d.m[0][1] = sin;
	d.m[1][0] = -sin; d.m[1][1] = cos;
	multiply(d);
}

void CMatrix::lookAt(const CVector& position, const CVector& target, const CVector& up) {
	CVector zaxis(target.x - position.x,
		target.y - position.y, target.z - position.z);
	zaxis.normalize();
	CVector xaxis = CVector::cross(up, zaxis);
	xaxis.normalize();
	CVector yaxis = CVector::cross(zaxis, xaxis);
	m[0][0] = xaxis.x;
	m[0][1] = yaxis.x;
	m[0][2] = zaxis.x;
	m[0][3] = 0.0f;
	m[1][0] = xaxis.y;
	m[1][1] = yaxis.y;
	m[1][2] = zaxis.y;
	m[1][3] = 0.0f;
	m[2][0] = xaxis.z;
	m[2][1] = yaxis.z;
	m[2][2] = zaxis.z;
	m[2][3] = 0.0f;
	m[3][0] = -CVector::dot(xaxis, position);
	m[3][1] = -CVector::dot(yaxis, position);
	m[3][2] = -CVector::dot(zaxis, position);
	m[3][3] = 1.0f;
}

void CMatrix::projection(float fov, float aspect, float zNear, float zFar) {
	float h = 1.0f / tanf((float)M_PI * fov * 0.5f);
	float w = h / aspect;
	identity();
	m[0][0] = w;
	m[1][1] = h;
	m[2][2] = zFar / (zFar - zNear);
	m[2][3] = 1.0f;
	m[3][2] = -zNear * zFar / (zFar - zNear);
	m[3][3] = 0.0f;
}

void CMatrix::viewport(int w, int h) {
	identity();
	m[0][0] = (float)(w / 2);
	m[1][1] = (float)(-h / 2);
	m[3][0] = (float)(w / 2);
	m[3][1] = (float)(h / 2);
}

bool CMatrix::transform(int* pX, int* pY, int* pZ, int* pRHW, const VERTEX& v) {
	float x = v.x * m[0][0] + v.y * m[1][0] + v.z * m[2][0] + m[3][0];
	float y = v.x * m[0][1] + v.y * m[1][1] + v.z * m[2][1] + m[3][1];
	float z = v.x * m[0][2] + v.y * m[1][2] + v.z * m[2][2] + m[3][2];
	float w = v.x * m[0][3] + v.y * m[1][3] + v.z * m[2][3] + m[3][3];
	if (w == 0.0f) {
		return false;	// 0割ガード
	}
	*pX = (int)(x / w);
	*pY = (int)(y / w);
	*pZ = (int)(z / w * zFar);	// zFarは0.0f〜1.0fまでの範囲を0〜zFarに広げる
	*pRHW = (int)(512.0f / w);	// 512.0fは精度確保のための値
								// この値を大きくすればテクスチャマッピングの精度は上がるが
								// CScan::linear()のループ回数が増えるので遅くなる
	if (*pRHW < 0) {
		return false;	// アンダーフローガード
	}
	return true;
}
